from past.builtins import basestring
import os
from itertools import chain
import os.path as op
import hashlib
import pydicom
from arcana.data.file_format import FileFormat
from arcana.data.file_format.standard import directory_format
from arcana.utils import split_extension, parse_value
from arcana.exceptions import (
    ArcanaError, ArcanaFileFormatError, ArcanaUsageError,
    ArcanaFileFormatNotRegisteredError, ArcanaNameError,
    ArcanaDataNotDerivedYetError)
from .base import BaseFileset, BaseField

DICOM_SERIES_NUMBER_TAG = ('0020', '0011')


class BaseItem(object):

    is_spec = False

    def __init__(self, subject_id, visit_id, repository, from_study,
                 exists, record):
        self._subject_id = subject_id
        self._visit_id = visit_id
        self._repository = repository
        self._from_study = from_study
        self._exists = exists
        self._record = record

    def __eq__(self, other):
        return (self.subject_id == other.subject_id and
                self.visit_id == other.visit_id and
                self.from_study == other.from_study and
                self.exists == other.exists and
                self._record == other._record)

    def __hash__(self):
        return (hash(self.subject_id) ^
                hash(self.visit_id) ^
                hash(self.from_study) ^
                hash(self.exists) ^
                hash(self._record))

    def find_mismatch(self, other, indent=''):
        sub_indent = indent + '  '
        mismatch = ''
        if self.subject_id != other.subject_id:
            mismatch += ('\n{}subject_id: self={} v other={}'
                         .format(sub_indent, self.subject_id,
                                 other.subject_id))
        if self.visit_id != other.visit_id:
            mismatch += ('\n{}visit_id: self={} v other={}'
                         .format(sub_indent, self.visit_id,
                                 other.visit_id))
        if self.from_study != other.from_study:
            mismatch += ('\n{}from_study: self={} v other={}'
                         .format(sub_indent, self.from_study,
                                 other.from_study))
        if self.exists != other.exists:
            mismatch += ('\n{}exists: self={} v other={}'
                         .format(sub_indent, self.exists,
                                 other.exists))
        if self._record != other._record:
            mismatch += ('\n{}_record: self={} v other={}'
                         .format(sub_indent, self._record,
                                 other._record))
        return mismatch

    @property
    def derived(self):
        return self.from_study is not None

    @property
    def repository(self):
        return self._repository

    @property
    def exists(self):
        return self._exists

    @property
    def subject_id(self):
        return self._subject_id

    @property
    def visit_id(self):
        return self._visit_id

    @property
    def session_id(self):
        return (self.subject_id, self.visit_id)

    @property
    def from_study(self):
        return self._from_study

    @property
    def record(self):
        return self._record

    @record.setter
    def record(self, record):
        if self.name not in record.outputs:
            raise ArcanaNameError(
                self.name,
                "{} was not found in outputs {} of provenance record".format(
                    self.name, record.outputs.keys(), record))
        self._record = record

    @property
    def recorded_checksums(self):
        if self.record is None:
            return None
        else:
            return self._record.outputs[self.name]

    def initkwargs(self):
        dct = super(Fileset, self).initkwargs()
        dct['repository'] = self.repository
        dct['subject_id'] = self.subject_id
        dct['visit_id'] = self.visit_id
        dct['from_study'] = self._from_study
        return dct


class Fileset(BaseItem, BaseFileset):
    """
    A representation of a fileset within the repository.

    Parameters
    ----------
    name : str
        The name of the fileset
    format : FileFormat
        The file format used to store the fileset.
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    derived : bool
        Whether the scan was generated or acquired. Depending on the repository
        used to store the fileset this is used to determine the location of the
        fileset.
    path : str | None
        The path to the fileset (for repositorys on the local system)
    id : int | None
        The ID of the fileset in the session. To be used to
        distinguish multiple filesets with the same scan type in the
        same session, e.g. scans taken before and after a task. For
        repositorys where this isn't stored (i.e. Local), id can be None
    subject_id : int | str | None
        The id of the subject which the fileset belongs to
    visit_id : int | str | None
        The id of the visit which the fileset belongs to
    repository : BaseRepository
        The repository which the fileset is stored
    from_study : str
        Name of the Arcana study that that generated the field
    exists : bool
        Whether the fileset exists or is just a placeholder for a derivative
    bids_attr : ??
        BIDS attributes
    checksums : dict[str, str]
        A checksums of all files within the fileset in a dictionary sorted by
        relative file paths
    record : arcana.pipeline.provenance.Record | None
        The provenance record for the pipeline that generated the file set,
        if applicable
    """

    def __init__(self, name, format=None, frequency='per_session', # @ReservedAssignment @IgnorePep8
                 path=None, id=None, uri=None, subject_id=None, # @ReservedAssignment @IgnorePep8
                 visit_id=None, repository=None, from_study=None,
                 exists=True, bids_attr=None, checksums=None,
                 record=None):
        BaseFileset.__init__(self, name=name, format=format,
                             frequency=frequency)
        BaseItem.__init__(self, subject_id, visit_id, repository,
                          from_study, exists, record)
        if path is not None:
            path = op.abspath(op.realpath(path))
        self._path = path
        self._uri = uri
        self._bids_attr = bids_attr
        if id is None and path is not None and format.name == 'dicom':
            self._id = int(self.dicom_values([DICOM_SERIES_NUMBER_TAG])
                           [DICOM_SERIES_NUMBER_TAG])
        else:
            self._id = id
        self._checksums = checksums

    def __eq__(self, other):
        eq = (BaseFileset.__eq__(self, other) and
              BaseItem.__eq__(self, other) and
              self.id == other.id and
              self.bids_attr == other.bids_attr and
              self.checksums == other.checksums)
        # Avoid having to cache fileset in order to test equality unless they
        # are already both cached
        try:
            if self._path is not None and other._path is not None:
                eq &= (self._path == other._path)
        except AttributeError:
            return False
        return eq

    def __hash__(self):
        return (BaseFileset.__hash__(self) ^
                BaseItem.__hash__(self) ^
                hash(self.id) ^
                hash(self.bids_attr) ^
                hash(self.checksums))

    def __lt__(self, other):
        if isinstance(self.id, int) and isinstance(other.id, basestring):
            return True
        elif isinstance(self.id, basestring) and isinstance(other.id, int):
            return False
        else:
            if self.id == other.id:
                # If ids are equal order depending on study name
                # with acquired (from_study==None) coming first
                if self.from_study is None:
                    return other.from_study is None
                elif other.from_study is None:
                    return False
                else:
                    return self.from_study < other.from_study
            else:
                return self.id < other.id

    def __repr__(self):
        return ("{}('{}', {}, '{}', subj={}, vis={}, stdy='{}', exists={})"
                .format(
                    type(self).__name__, self.name, self.format,
                    self.frequency, self.subject_id,
                    self.visit_id, self.from_study,
                    self.exists))

    @property
    def fname(self):
        return self.name + self.format.ext_str

    @property
    def bids_attr(self):
        return self._bids_attr

    def find_mismatch(self, other, indent=''):
        mismatch = BaseFileset.find_mismatch(self, other, indent)
        mismatch += BaseItem.find_mismatch(self, other, indent)
        sub_indent = indent + '  '
        if self._path != other._path:
            mismatch += ('\n{}path: self={} v other={}'
                         .format(sub_indent, self._path,
                                 other._path))
        if self._id != other._id:
            mismatch += ('\n{}id: self={} v other={}'
                         .format(sub_indent, self._id,
                                 other._id))
        if self._bids_attr != other._bids_attr:
            mismatch += ('\n{}bids_attr: self={} v other={}'
                         .format(sub_indent, self._bids_attr,
                                 other._bids_attr))
        if self.checksums != other.checksums:
            mismatch += ('\n{}checksum: self={} v other={}'
                         .format(sub_indent, self.checksums,
                                 other.checksums))
        return mismatch

    @property
    def path(self):
        if not self.exists:
            raise ArcanaDataNotDerivedYetError(
                self.name,
                "Cannot access path of {} as it hasn't been derived yet"
                .format(self))
        if self._path is None:
            if self.repository is not None:
                self.get()  # Retrieve from repository
            else:
                raise ArcanaError(
                    "Neither path nor repository has been set for Fileset("
                    "'{}')".format(self.name))
        return self._path

    @path.setter
    def path(self, path):
        if path is not None:
            path = op.abspath(op.realpath(path))
            self._exists = True
        self._path = path
        self._checksums = self.calculate_checksums()
        self.put()  # Push to repository

    @property
    def paths(self):
        """Iterates through all files in the set"""
        if self.format.directory:
            return chain(*((op.join(root, f) for f in files)
                           for root, _, files in os.walk(self.path)))
        else:
            # FIXME: Need to add support for headers/side-car files
            return iter([self.path])

    def basename(self, **kwargs):  # @UnusedVariable
        return self.name

    @property
    def id(self):
        if self._id is None:
            return self.name
        else:
            return self._id

    @id.setter
    def id(self, id):  # @ReservedAssignment
        if self._id is None:
            self._id = id
        elif id != self._id:
            raise ArcanaUsageError("Can't change value of ID for {} from {} "
                                   "to {}".format(self, self._id, id))

    @property
    def uri(self):
        return self._uri

    @uri.setter
    def uri(self, uri):
        if self._uri is None:
            self._uri = uri
        elif uri != self._uri:
            raise ArcanaUsageError("Can't change value of URI for {} from {} "
                                   "to {}".format(self, self._uri, uri))

    @property
    def checksums(self):
        if not self.exists:
            raise ArcanaDataNotDerivedYetError(
                self.name,
                "Cannot access checksums of {} as it hasn't been derived yet"
                .format(self))
        if self._checksums is None:
            if self.repository is not None:
                self._checksums = self.repository.get_checksums(self)
            if self._checksums is None:
                self._checksums = self.calculate_checksums()
        return self._checksums

    def calculate_checksums(self):
        checksums = {}
        for fpath in self.paths:
            with open(fpath, 'rb') as f:
                checksums[op.relpath(fpath, self.path)] = hashlib.md5(
                    f.read()).hexdigest()
        return checksums

    @classmethod
    def from_path(cls, path, frequency='per_session', format=None,  # @ReservedAssignment @IgnorePep8
                  **kwargs):
        if not op.exists(path):
            raise ArcanaUsageError(
                "Attempting to read Fileset from path '{}' but it "
                "does not exist".format(path))
        if op.isdir(path):
            within_exts = frozenset(
                split_extension(f)[1] for f in os.listdir(path)
                if not f.startswith('.'))
            if format is None:
                # Try to guess format
                try:
                    format = FileFormat.by_within_dir_exts(within_exts)  # @ReservedAssignment @IgnorePep8
                except ArcanaFileFormatNotRegisteredError:
                    # Fall back to general directory format
                    format = directory_format  # @ReservedAssignment
            name = op.basename(path)
        else:
            filename = op.basename(path)
            name, ext = split_extension(filename)
            if format is None:
                try:
                    format = FileFormat.by_ext(ext)  # @ReservedAssignment @IgnorePep8
                except ArcanaFileFormatNotRegisteredError as e:
                    raise ArcanaFileFormatNotRegisteredError(
                        str(e) + ", which is required to identify the "
                        "format of the fileset at '{}'".format(path))
        return cls(name, format, frequency=frequency,
                   path=path, **kwargs)

    def dicom(self, index):
        """
        Returns a PyDicom object for the DICOM file at index 'index'

        Parameters
        ----------
        fileset : Fileset
            The fileset to read a DICOM file from
        index : int
            The index of the DICOM file in the fileset to read

        Returns
        -------
        dcm : pydicom.DICOM
            A PyDicom file object
        """
        if self.format.name != 'dicom':
            raise ArcanaFileFormatError(
                "Can not read DICOM header as {} is not in DICOM format"
                .format(self))
        fnames = sorted([f for f in os.listdir(self.path)
                         if not f.startswith('.')])
        with open(op.join(self.path, fnames[index]), 'rb') as f:
            dcm = pydicom.dcmread(f)
        return dcm

    def dicom_values(self, tags):
        """
        Returns a dictionary with the DICOM header fields corresponding
        to the given tag names

        Parameters
        ----------
        tags : List[Tuple[str, str]]
            List of DICOM tag values as 2-tuple of strings, e.g.
            [('0080', '0020')]
        repository_login : <repository-login-object>
            A login object for the repository to avoid having to relogin
            for every dicom_header call.

        Returns
        -------
        dct : Dict[Tuple[str, str], str|int|float]
        """
        try:
            if (self._path is None and self._repository is not None and
                    hasattr(self.repository, 'dicom_header')):
                hdr = self.repository.dicom_header(self)
                dct = {t: hdr[t] for t in tags}
            else:
                # Get the DICOM object for the first file in the fileset
                dcm = self.dicom(0)
                dct = {t: dcm[t].value for t in tags}
        except KeyError as e:
            raise ArcanaNameError(
                e.args[0], "{} does not have dicom tag {}".format(
                    self, e.args[0]))
        return dct

    def initkwargs(self):
        dct = BaseFileset.initkwargs(self)
        dct.update(BaseItem.initkwargs(self))
        dct['path'] = self.path
        dct['id'] = self.id
        dct['uri'] = self.uri
        dct['bids_attr'] = self.bids_attr
        dct['checksums'] = self.checksums
        return dct

    def get(self):
        if self.repository is not None:
            self._exists = True
            self._path = self.repository.get_fileset(self)

    def put(self):
        if self.repository is not None and self._path is not None:
            self.repository.put_fileset(self)

    def get_array(self):
        return self.format.get_array(self.path)

    def get_header(self):
        return self.format.get_header(self.path)


class Field(BaseItem, BaseField):
    """
    A representation of a value field in the repository.

    Parameters
    ----------
    name : str
        The name of the fileset
    dtype : type
        The datatype of the value. Can be one of (float, int, str)
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    derived : bool
        Whether or not the value belongs in the derived session or not
    subject_id : int | str | None
        The id of the subject which the field belongs to
    visit_id : int | str | None
        The id of the visit which the field belongs to
    repository : BaseRepository
        The repository which the field is stored
    from_study : str
        Name of the Arcana study that that generated the field
    exists : bool
        Whether the field exists or is just a placeholder for a derivative
    record : arcana.pipeline.provenance.Record | None
        The provenance record for the pipeline that generated the field,
        if applicable
    """

    def __init__(self, name, value=None, dtype=None,
                 frequency='per_session', array=None, subject_id=None,
                 visit_id=None, repository=None, from_study=None,
                 exists=True, record=None):
        # Try to determine dtype and array from value if they haven't
        # been provided.
        if value is None:
            if dtype is None:
                raise ArcanaUsageError(
                    "Either 'value' or 'dtype' must be provided to "
                    "Field init")
            if array is None:
                raise ArcanaUsageError(
                    "Either 'value' or 'array' must be provided to "
                    "Field init")
        else:
            value = parse_value(value)
            if isinstance(value, list):
                if array is False:
                    raise ArcanaUsageError(
                        "Array value passed to '{}', which is explicitly not "
                        "an array ({})".format(name, value))
                array = True
            else:
                if array:
                    raise ArcanaUsageError(
                        "Non-array value ({}) passed to '{}', which expects "
                        "array{}".format(value, name,
                                         ('of type {}'.format(dtype)
                                          if dtype is not None else '')))
                array = False
            if dtype is None:
                if array:
                    dtype = type(value[0])
                else:
                    dtype = type(value)
            else:
                # Ensure everything is cast to the correct type
                if array:
                    value = [dtype(v) for v in value]
                else:
                    value = dtype(value)
        BaseField.__init__(self, name, dtype, frequency, array)
        BaseItem.__init__(self, subject_id, visit_id, repository,
                          from_study, exists, record)
        self._value = value

    def __eq__(self, other):
        return (BaseField.__eq__(self, other) and
                BaseItem.__eq__(self, other) and
                self.value == other.value)

    def __hash__(self):
        return (BaseField.__hash__(self) ^
                BaseItem.__hash__(self) ^
                hash(self.value))

    def find_mismatch(self, other, indent=''):
        mismatch = BaseField.find_mismatch(self, other, indent)
        mismatch += BaseItem.find_mismatch(self, other, indent)
        sub_indent = indent + '  '
        if self.value != other.value:
            mismatch += ('\n{}value: self={} v other={}'
                         .format(sub_indent, self.value,
                                 other.value))
        return mismatch

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __str__(self):
        if self.array:
            val = '[' + ','.join(self._to_str(v) for v in self.value) + ']'
        else:
            val = self._to_str(self.value)
        return val

    def _to_str(self, val):
        if self.dtype is str:
            val = '"{}"'.format(val)
        else:
            val = str(val)
        return val

    def __lt__(self, other):
        if self.name == other.name:
            # If ids are equal order depending on study name
            # with acquired (from_study==None) coming first
            if self.from_study is None:
                return other.from_study is None
            elif other.from_study is None:
                return False
            else:
                return self.from_study < other.from_study
        else:
            return self.name < other.name

    def __repr__(self):
        return ("{}('{}',{} '{}', subj={}, vis={}, stdy='{}', exists={})"
                .format(
                    type(self).__name__, self.name,
                    ("{}, ".format(self._value)
                     if self._value is not None else ''),
                    self.frequency, self.subject_id,
                    self.visit_id, self.from_study,
                    self.exists))

    def basename(self, **kwargs):  # @UnusedVariable
        return self.name

    @property
    def value(self):
        if not self.exists:
            raise ArcanaDataNotDerivedYetError(
                self.name,
                "Cannot access value of {} as it hasn't been "
                "derived yet".format(repr(self)))
        if self._value is None:
            if self.repository is not None:
                self._value = self.repository.get_field(self)
            else:
                raise ArcanaError(
                    "Neither value nor repository has been set for Field("
                    "'{}')".format(self.name))
        return self._value

    @value.setter
    def value(self, value):
        if self.array:
            self._value = [self.dtype(v) for v in value]
        else:
            self._value = self.dtype(value)
        self._exists = True
        self.put()

    @property
    def checksums(self):
        """
        For duck-typing with filesets in checksum management. Instead of a
        checksum, just the value of the field is used
        """
        return self.value

    def initkwargs(self):
        dct = BaseField.initkwargs(self)
        dct.update(BaseItem.initkwargs(self))
        dct['value'] = self.value
        return dct

    def get(self):
        if self.repository is not None:
            self._exists = True
            self._value = self.repository.get_field(self)

    def put(self):
        if self.repository is not None and self._value is not None:
            self.repository.put_field(self)



from pprint import pformat
import os.path as op
import tempfile
from copy import copy
import shutil
from nipype.interfaces.utility import Merge, Split  # @IgnorePep8
from arcana.utils.testing import (
    BaseTestCase, BaseMultiSubjectTestCase, TestMath)  # @IgnorePep8
from arcana.processor import LinearProcessor
from arcana.study.base import Study, StudyMetaClass  # @IgnorePep8
from arcana.study.parameter import ParameterSpec, SwitchSpec  # @IgnorePep8
from arcana.data import (
    AcquiredFilesetSpec, FilesetSpec, FieldSpec, FilesetSelector,
    AcquiredFieldSpec, FieldSelector)  # @IgnorePep8
from arcana.data.file_format.standard import text_format  # @IgnorePep8
from future.utils import with_metaclass  # @IgnorePep8
from arcana.pipeline.provenance import Record
from arcana.data import Fileset, Field
from arcana.repository import Tree
from arcana.environment import BaseRequirement
from arcana.exceptions import (
    ArcanaReprocessException, ArcanaProtectedOutputConflictError)


class DummyRequirement(BaseRequirement):

    def __init__(self, *args, **kwargs):
        self.detected_version = kwargs.pop('detected_version')
        super(DummyRequirement, self).__init__(*args, **kwargs)

    def detect_version_str(self):
        return self.detected_version


a_req = DummyRequirement('a', detected_version='1.0.1')
b_req = DummyRequirement('b', detected_version='2.7.2')
c_req = DummyRequirement('c', detected_version='0.2a2')
d_req = DummyRequirement('d', detected_version='0.9.dev10')


class TestProvStudy(with_metaclass(StudyMetaClass, Study)):

    add_data_specs = [
        AcquiredFilesetSpec('acquired_fileset1', text_format),
        AcquiredFilesetSpec('acquired_fileset2', text_format),
        AcquiredFilesetSpec('acquired_fileset3', text_format),
        AcquiredFieldSpec('acquired_field1', float),
        FilesetSpec('derived_fileset1', text_format, 'pipeline2'),
        FieldSpec('derived_field1', float, 'pipeline1', array=True),
        FieldSpec('derived_field2', float, 'pipeline3'),
        FieldSpec('derived_field3', float, 'pipeline3'),
        FieldSpec('derived_field4', float, 'pipeline2')]

    add_param_specs = [
        SwitchSpec('extra_req', False),
        SwitchSpec('branch', 'foo', ('foo', 'bar', 'wee')),
        ParameterSpec('multiplier', 10.0),
        ParameterSpec('subtract', 3)]

    def pipeline1(self, **name_maps):
        pipeline = self.new_pipeline(
            'pipeline1',
            desc="",
            references=[],
            name_maps=name_maps)
        math1 = pipeline.add(
            'math1',
            TestMath(
                op='add'),
            inputs={
                'x': ('acquired_fileset1', text_format),
                'y': ('acquired_fileset2', text_format)},
            requirements=[
                a_req.v('1.0.1'),
                b_req.v(2)])
        math2 = pipeline.add(
            'math2',
            TestMath(
                op='add'),
            inputs={
                'y': ('acquired_field1', float)},
            connect={
                'x': (math1, 'z')},
            requirements=[
                c_req.v(0.1)])
        # Set up different requirements based on switch
        math3_reqs = [a_req.v(1)]
        if self.branch('extra_req'):
            math3_reqs.append(d_req.v('0.8.6'))
        math3 = pipeline.add(
            'math3',
            TestMath(
                op='mul',
                y=self.parameter('multiplier')),
            connect={
                'x': (math2, 'z')},
            requirements=[
                b_req.v('2.7.0', '3.0')])
        pipeline.add(
            'merge1',
            Merge(3),
            connect={
                'in1': (math1, 'z'),
                'in2': (math2, 'z'),
                'in3': (math3, 'z')},
            outputs={
                'out': ('derived_field1', float)},
            requirements=math3_reqs)
        return pipeline

    def pipeline2(self, **name_maps):
        pipeline = self.new_pipeline(
            'pipeline2',
            desc="",
            references=[],
            name_maps=name_maps)
        split = pipeline.add(
            'split',
            Split(
                splits=[1, 1, 1],
                squeeze=True),
            inputs={
                'inlist': ('derived_field1', float)})
        math1 = pipeline.add(
            'math1',
            TestMath(
                op='add',
                as_file=True),
            inputs={
                'y': ('acquired_fileset3', text_format)},
            connect={
                'x': (split, 'out3')},
            requirements=[
                a_req.v('1.0')])
        math2 = pipeline.add(
            'math2',
            TestMath(
                op='add',
                as_file=True),
            inputs={
                'y': ('acquired_field1', float)},
            connect={
                'x': (math1, 'z')},
            outputs={
                'z': ('derived_fileset1', text_format)},
            requirements=[
                c_req.v(0.1)])
        pipeline.add(
            'math3',
            TestMath(
                op='sub',
                as_file=False,
                y=-1),
            connect={
                'x': (math2, 'z')},
            outputs={
                'z': ('derived_field4', float)})
        return pipeline

    def pipeline3(self, **name_maps):
        pipeline = self.new_pipeline(
            'pipeline3',
            desc="",
            references=[],
            name_maps=name_maps)
        pipeline.add(
            'math1',
            TestMath(
                op='add',
                as_file=False),
            inputs={
                'x': ('acquired_fileset2', text_format),
                'y': ('derived_fileset1', text_format)},
            outputs={
                'z': ('derived_field2', float)},
            requirements=[
                a_req.v('1.0')])
        pipeline.add(
            'math2',
            TestMath(
                op='sub',
                as_file=False,
                y=self.parameter('subtract')),
            inputs={
                'x': ('acquired_field1', float)},
            outputs={
                'z': ('derived_field3', float)},
            requirements=[
                c_req.v(0.1)])
        return pipeline


class TestProvStudyAddNode(with_metaclass(StudyMetaClass, TestProvStudy)):

    add_param_specs = [
        ParameterSpec('a', 50.0),
        ParameterSpec('b', 25.0)]

    def pipeline2(self, **name_maps):
        pipeline = super(TestProvStudyAddNode, self).pipeline2(**name_maps)
        pipeline.add(
            'math4',
            TestMath(
                op='mul',
                as_file=True,
                x=self.parameter('a'),
                y=self.parameter('b')),
            outputs={
                'z': ('derived_fileset1', text_format)})
        return pipeline


class TestProvStudyAddConnect(with_metaclass(StudyMetaClass, TestProvStudy)):

    def pipeline1(self, **name_maps):
        pipeline = super(TestProvStudyAddConnect, self).pipeline1(**name_maps)
        pipeline.connect_input('acquired_field1', pipeline.node('math3'), 'y',
                               float)
        return pipeline


def change_value_w_prov(field, new_value):
    """
    Changes the value of a field and updates the corresponding provenance
    record so that it appears if the value was given by the pipeline. Used
    to detect whether a value has been reprocessed unnecessarily, as in that
    case the value will be overwritten to the value actually outputted by
    the pipeline
    """
    # Change value in repository to test that the pipeline isn't rerun
    field.value = new_value
    # Update provenance record so it isn't determined to be protected
    record = field.record
    record.prov['outputs'][field.name] = new_value
    field.repository.put_record(record)


STUDY_INPUTS = [FilesetSelector('acquired_fileset1', text_format,
                                'acquired_fileset1'),
                FilesetSelector('acquired_fileset2', text_format,
                                'acquired_fileset2'),
                FilesetSelector('acquired_fileset3', text_format,
                                'acquired_fileset3'),
                FieldSelector('acquired_field1', int, 'acquired_field1')]


INPUT_DATASETS = {'acquired_fileset1': '1.0',
                  'acquired_fileset2': '2.0',
                  'acquired_fileset3': '3.0'}

INPUT_FIELDS = {'acquired_field1': 11}


class TestProvBasic(BaseTestCase):

    INPUT_DATASETS = INPUT_DATASETS
    INPUT_FIELDS = INPUT_FIELDS

    def test_altered_workflow(self):
        """
        Tests whether data is regenerated if the pipeline workflows are altered
        """
        study_name = 'add_node'
        # Test vanilla study
        study = self.create_study(
            TestProvStudy,
            study_name,
            inputs=STUDY_INPUTS)
        self.assertEqual(study.data('derived_field2').value(*self.SESSION),
                         156.0)
        # Rerun results of altered study
        study = self.create_study(
            TestProvStudyAddNode,
            study_name,
            processor=LinearProcessor(self.work_dir, reprocess=True),
            inputs=STUDY_INPUTS)
        self.assertEqual(study.data('derived_field2').value(*self.SESSION),
                         1252.0)
        study_name = 'add_connect'
        # Test vanilla study
        study = self.create_study(
            TestProvStudy,
            study_name,
            inputs=STUDY_INPUTS)
        self.assertEqual(study.data('derived_field2').value(*self.SESSION),
                         156.0)
        # Rerun results of altered study
        study = self.create_study(
            TestProvStudyAddConnect,
            study_name,
            processor=LinearProcessor(self.work_dir, reprocess=True),
            inputs=STUDY_INPUTS)
        self.assertEqual(study.data('derived_field2').value(*self.SESSION),
                         170.0)

    def test_unchanged_workflow(self):
        """
        Tests that when a parameter is changed that doesn't effect the
        workflows that generated a value, then the data isn't regenerated
        """
        study_name = 'changed_parameter'
        new_value = -99.0
        # Test vanilla study
        study = self.create_study(
            TestProvStudy,
            study_name,
            inputs=STUDY_INPUTS)
        derived_field4 = study.data('derived_field4').item(*self.SESSION)
        self.assertEqual(derived_field4.value, 155.0)
        # Change value to a new value to see if it gets overwritten even
        # it shouldn't as the parameter that is changed doesn't impact on
        # derived field 4.
        change_value_w_prov(derived_field4, new_value)
        study = self.create_study(
            TestProvStudy,
            study_name,
            inputs=STUDY_INPUTS,
            parameters={'subtract': 100})
        new_derived_field4 = study.data('derived_field4').item(*self.SESSION)
        self.assertEqual(new_derived_field4.value, new_value)
        self.assertEqual(
            new_derived_field4.record.prov['outputs']['derived_field4'],
            new_value)

    def test_protect_manually(self):
        """Protect manually altered files and fields from overwrite"""
        study_name = 'manual_protect'
        protected_derived_field4_value = -99.0
        protected_derived_fileset1_value = -999.0
        # Test vanilla study
        study = self.create_study(
            TestProvStudy,
            study_name,
            inputs=STUDY_INPUTS)
        derived_fileset1_collection, derived_field4_collection = study.data(
            ('derived_fileset1', 'derived_field4'))
        self.assertContentsEqual(derived_fileset1_collection, 154.0)
        self.assertEqual(derived_field4_collection.value(*self.SESSION), 155.0)
        # Rerun with new parameters
        study = self.create_study(
            TestProvStudy,
            study_name,
            inputs=STUDY_INPUTS,
            processor=LinearProcessor(self.work_dir, reprocess=True),
            parameters={'multiplier': 100.0})
        derived_fileset1_collection, derived_field4_collection = study.data(
            ('derived_fileset1', 'derived_field4'))
        self.assertContentsEqual(derived_fileset1_collection, 1414.0)
        derived_field4 = derived_field4_collection.item(*self.SESSION)
        self.assertEqual(derived_field4.value, 1415.0)
        # Manually changing the value (or file contents) of a derivative value
        # (without also altering the saved provenance record) will mean
        # that new value/file will be "protected" from reprocessing, and will
        # need to be deleted in order to be regenerated
        derived_field4.value = protected_derived_field4_value
        # Since derived_fileset1 needs to be reprocessed but
        study = self.create_study(
            TestProvStudy,
            study_name,
            processor=LinearProcessor(self.work_dir, reprocess=True),
            inputs=STUDY_INPUTS,
            parameters={'multiplier': 1000.0})
        # Check to see protected conflict error is raise if only one of
        # derived field4/fileset1 is protected
        self.assertRaises(
            ArcanaProtectedOutputConflictError,
            study.data,
            ('derived_fileset1', 'derived_field4'))
        with open(derived_fileset1_collection.path(*self.SESSION), 'w') as f:
            f.write(str(protected_derived_fileset1_value))
        study.clear_caches()
        # Protect the output of derived_fileset1 as well and it should return
        # the protected values
        derived_fileset1_collection, derived_field4_collection = study.data(
            ('derived_fileset1', 'derived_field4'))
        self.assertContentsEqual(derived_fileset1_collection,
                                 protected_derived_fileset1_value)
        self.assertEqual(derived_field4_collection.value(*self.SESSION),
                         protected_derived_field4_value)


class TestProvInputChange(BaseTestCase):

    INPUT_DATASETS = INPUT_DATASETS
    INPUT_FIELDS = INPUT_FIELDS

    def test_input_change(self):
        study_name = 'input_change_study'
        study = self.create_study(
            TestProvStudy,
            study_name,
            inputs=STUDY_INPUTS)
        self.assertEqual(study.data('derived_field2').value(*self.SESSION),
                         156.0)
        # Change acquired file contents, which should cause the checksum check
        # to fail
        with open(study.data('acquired_fileset1').path(*self.SESSION),
                  'w') as f:
            f.write('99.9')
        # Should detect that the input has changed and throw an error
        self.assertRaises(
            ArcanaReprocessException,
            study.data,
            'derived_field2')
        new_study = self.create_study(
            TestProvStudy,
            study_name,
            processor=LinearProcessor(self.work_dir, reprocess=True),
            inputs=STUDY_INPUTS)
        self.assertEqual(
            new_study.data('derived_field2').value(*self.SESSION), 1145.0)


class TestProvDialationStudy(with_metaclass(StudyMetaClass, Study)):

    add_data_specs = [
        AcquiredFieldSpec('acquired_field1', int),
        FieldSpec('derived_field1', int, 'pipeline1'),
        FieldSpec('derived_field2', int, 'pipeline2',
                  frequency='per_subject'),
        FieldSpec('derived_field3', int, 'pipeline3',
                  frequency='per_visit'),
        FieldSpec('derived_field4', int, 'pipeline4',
                  frequency='per_study'),
        FieldSpec('derived_field5', int, 'pipeline5')]

    add_param_specs = [
        ParameterSpec('increment', 1),
        ParameterSpec('pipeline3_op', 'add')]

    def pipeline1(self, **name_maps):
        pipeline = self.new_pipeline(
            'pipeline1',
            desc="",
            references=[],
            name_maps=name_maps)
        pipeline.add(
            'math',
            TestMath(
                op='add',
                y=self.parameter('increment'),
                as_file=False),
            inputs={
                'x': ('acquired_field1', int)},
            outputs={
                'z': ('derived_field1', int)})
        return pipeline

    def pipeline2(self, **name_maps):
        pipeline = self.new_pipeline(
            'pipeline2',
            desc="",
            references=[],
            name_maps=name_maps)
        pipeline.add(
            'math',
            TestMath(
                op='add',
                as_file=False),
            inputs={
                'x': ('derived_field1', int)},
            outputs={
                'z': ('derived_field2', int)},
            joinsource=self.VISIT_ID,
            joinfield='x')
        return pipeline

    def pipeline3(self, **name_maps):
        pipeline = self.new_pipeline(
            'pipeline3',
            desc="",
            references=[],
            name_maps=name_maps)
        pipeline.add(
            'math',
            TestMath(
                op=self.parameter('pipeline3_op'),
                as_file=False),
            inputs={
                'x': ('derived_field1', int)},
            outputs={
                'z': ('derived_field3', int)},
            joinsource=self.SUBJECT_ID,
            joinfield='x')
        return pipeline

    def pipeline4(self, **name_maps):
        pipeline = self.new_pipeline(
            'pipeline4',
            desc="",
            references=[],
            name_maps=name_maps)
        merge1 = pipeline.add(
            'merge1',
            Merge(
                numinputs=1,
                ravel_inputs=True),
            inputs={
                'in1': ('derived_field1', int)},
            joinsource=self.SUBJECT_ID,
            joinfield='in1')
        merge2 = pipeline.add(
            'merge2',
            Merge(
                numinputs=1,
                ravel_inputs=True),
            connect={
                'in1': (merge1, 'out')},
            joinsource=self.VISIT_ID,
            joinfield='in1')
        pipeline.add(
            'math',
            TestMath(
                op='add',
                as_file=False),
            connect={
                'x': (merge2, 'out')},
            outputs={
                'z': ('derived_field4', int)})
        return pipeline

    def pipeline5(self, **name_maps):
        pipeline = self.new_pipeline(
            'pipeline5',
            desc="",
            references=[],
            name_maps=name_maps)
        merge = pipeline.add(
            'merge',
            Merge(
                numinputs=3),
            inputs={
                'in1': ('derived_field2', int),
                'in2': ('derived_field3', int),
                'in3': ('derived_field4', int)})
        pipeline.add(
            'math',
            TestMath(
                op='add',
                as_file=False),
            connect={
                'x': (merge, 'out')},
            outputs={
                'z': ('derived_field5', float)})
        return pipeline


class TestProvDialation(BaseMultiSubjectTestCase):
    """
    Tests the "dialation" of the to process array in the case that there are
    summary outputs (i.e. frequency != 'per_session') of a pipeline.
    """

    NUM_SUBJECTS = 2
    NUM_VISITS = 2
    STUDY_INPUTS = [FieldSelector('acquired_field1', int, 'acquired_field1')]

    DEFAULT_FIELD5_VALUES = {
        ('0', '0'): 41,
        ('0', '1'): 43,
        ('1', '0'): 61,
        ('1', '1'): 63}

    @property
    def input_tree(self):
        fields = []
        for subj_i in range(self.NUM_SUBJECTS):
            for visit_i in range(self.NUM_VISITS):
                fields.append(
                    Field(name='acquired_field1', value=visit_i + subj_i * 10,
                          dtype=int, frequency='per_session',
                          subject_id=str(subj_i), visit_id=str(visit_i)))
        return Tree.construct(fields=fields)

    def test_filter_dialation1(self):
        study_name = 'process_dialation'
        study = self.create_study(
            TestProvDialationStudy,
            study_name,
            inputs=self.STUDY_INPUTS)
        field2 = study.data(
            'derived_field2',
            session_ids=[('0', '0'),
                         ('1', '1')])
        self.assertEqual(field2.value(subject_id='0'), 3)
        self.assertEqual(field2.value(subject_id='1'), 23)
        field3 = study.data(
            'derived_field3',
            session_ids=[('0', '0'),
                         ('1', '1')])
        self.assertEqual(field3.value(visit_id='0'), 12)
        self.assertEqual(field3.value(visit_id='1'), 14)
        field4 = study.data(
            'derived_field4',
            session_ids=[('1', '1')])
        self.assertEqual(field4.value(), 26)

    def test_filter_dialation2(self):
        study_name = 'filter_dialation2'
        study = self.create_study(
            TestProvDialationStudy,
            study_name,
            inputs=self.STUDY_INPUTS)
        field5 = study.data(
            'derived_field5',
            session_ids=[('1', '1')])
        self.assertEqual(len(field5), 1)
        self.assertEqual(field5.value(subject_id='1', visit_id='1'),
                         self.DEFAULT_FIELD5_VALUES[('1', '1')])

        # Check that no more fields were generated than necessary
        tree = study.tree
        sess00 = study.tree.session(subject_id='0', visit_id='0')
        sess01 = study.tree.session(subject_id='0', visit_id='1')
        sess10 = study.tree.session(subject_id='1', visit_id='0')
        sess11 = study.tree.session(subject_id='1', visit_id='1')
        subj0 = study.tree.subject('0')
        subj1 = study.tree.subject('1')
        vis0 = study.tree.visit('0')
        vis1 = study.tree.visit('1')
        self.assertEqual(list(tree.field_keys),
                         [('derived_field4', study_name)])
        self.assertFalse(list(subj0.field_keys))
        self.assertEqual(list(subj1.field_keys),
                         [('derived_field2', study_name)])
        self.assertFalse(list(vis0.field_keys))
        self.assertEqual(list(vis1.field_keys),
                         [('derived_field3', study_name)])
        self.assertEqual(sorted(sess00.field_keys),
                         [('acquired_field1', None),
                          ('derived_field1', study_name)])
        self.assertEqual(sorted(sess01.field_keys),
                         [('acquired_field1', None),
                          ('derived_field1', study_name)])
        self.assertEqual(sorted(sess10.field_keys),
                         [('acquired_field1', None),
                          ('derived_field1', study_name)])
        self.assertEqual(sorted(sess11.field_keys),
                         [('acquired_field1', None),
                          ('derived_field1', study_name),
                          ('derived_field5', study_name)])

    def test_process_dialation(self):
        study_name = 'process_dialation'
        new_value = -101
        study = self.create_study(
            TestProvDialationStudy,
            study_name,
            inputs=self.STUDY_INPUTS)
        study.data('derived_field5')

        def values_equal(field_name, values):
            for subj_i in range(self.NUM_SUBJECTS):
                for vis_i in range(self.NUM_VISITS):
                    sess = study.tree.session(subj_i, vis_i)
                    field = sess.field(field_name, from_study=study_name)
                    self.assertEqual(field.value,
                                     values[(str(subj_i), str(vis_i))])
        # Test generated values
        values_equal('derived_field5', self.DEFAULT_FIELD5_VALUES)
        # Tag the field 1 value so we can detect if it gets regenerated
        orig_field1_values = {}
        orig_field3_values = {}
        for vis_i in range(self.NUM_VISITS):
            for subj_i in range(self.NUM_SUBJECTS):
                sess = study.tree.session(subj_i, vis_i)
                field1 = sess.field('derived_field1', from_study=study_name)
                orig_field1_values[(str(subj_i),
                                    str(vis_i))] = field1.value
                change_value_w_prov(field1, new_value)
            field3 = study.tree.visit(vis_i).field('derived_field3',
                                                   from_study=study_name)
            orig_field3_values[str(vis_i)] = field3.value
        # Rerun analysis with new parameters
        study = self.create_study(
            TestProvDialationStudy,
            study_name,
            inputs=self.STUDY_INPUTS,
            processor=LinearProcessor(self.work_dir, reprocess=True),
            parameters={
                'pipeline3_op': 'mul'})
        study.data('derived_field3', subject_id='0', visit_id='0')
        values_equal('derived_field1',
                     {k: new_value for k in orig_field1_values})
        self.assertEqual(
            study.tree.visit('0').field('derived_field3',
                                        from_study=study_name).value,
            10201)
        self.assertEqual(
            study.tree.visit('1').field('derived_field3',
                                        from_study=study_name).value,
            orig_field3_values['1'])
        study = self.create_study(
            TestProvDialationStudy,
            study_name,
            inputs=self.STUDY_INPUTS,
            processor=LinearProcessor(self.work_dir, reprocess=True),
            parameters={
                'increment': 2})
        study.data('derived_field5', subject_id='0', visit_id='0')
        values_equal('derived_field1',
                     {k: v + 1 for k, v in orig_field1_values.items()})

    def test_dialation_protection(self):
        study_name = 'dialation_protection'
        study = self.create_study(
            TestProvDialationStudy,
            study_name,
            inputs=self.STUDY_INPUTS)
        field5 = study.data('derived_field5')
        for item in field5:
            self.assertEqual(item.value,
                             self.DEFAULT_FIELD5_VALUES[(item.subject_id,
                                                         item.visit_id)])
        field1 = study.data('derived_field1')
        field2 = study.data('derived_field2')
        field1.item(subject_id='0', visit_id='1').value = 1000000
        field1.item(subject_id='1', visit_id='1').value = 2000000
        # Manually change value of field 2
        field2.item(subject_id='0').value = -1000
        study = self.create_study(
            TestProvDialationStudy,
            study_name,
            processor=LinearProcessor(self.work_dir, reprocess=True),
            inputs=self.STUDY_INPUTS,
            parameters={
                'increment': 2})
        # Recalculate value of field5 with new field2 value
        field1, field2, field3, field4, field5 = study.data(
            ['derived_field1', 'derived_field2', 'derived_field3',
             'derived_field4', 'derived_field5'])
        self.assertEqual(field1.value(subject_id='0', visit_id='0'), 2)
        self.assertEqual(field1.value(subject_id='0', visit_id='1'), 1000000)
        self.assertEqual(field1.value(subject_id='1', visit_id='0'), 12)
        self.assertEqual(field1.value(subject_id='1', visit_id='1'), 2000000)
        self.assertEqual(field2.value(subject_id='0'), -1000)
        self.assertEqual(field2.value(subject_id='1'), 2000012)
        self.assertEqual(field3.value(visit_id='0'), 14)
        self.assertEqual(field3.value(visit_id='1'), 3000000)
        self.assertEqual(field4.value(), 3000014)
        self.assertEqual(field5.value(subject_id='0', visit_id='0'),
                         -1000 + 14 + 3000014)
        self.assertEqual(field5.value(subject_id='0', visit_id='1'),
                         -1000 + 3000000 + 3000014)
        self.assertEqual(field5.value(subject_id='1', visit_id='0'),
                         2000012 + 14 + 3000014)
        self.assertEqual(field5.value(subject_id='1', visit_id='1'),
                         2000012 + 3000000 + 3000014)



from builtins import object
import re
from arcana.utils import ExitStack
from copy import copy
from itertools import chain
from arcana.exceptions import (
    ArcanaUsageError, ArcanaFilesetSelectorError)
from .base import BaseFileset, BaseField
from .collection import FilesetCollection, FieldCollection


class BaseSelector(object):
    """
    Base class for Fileset and Field Selector classes
    """

    def __init__(self, pattern, is_regex, order, from_study,
                 fallback_to_default=False, repository=None, study_=None,
                 collection_=None):
        self._pattern = pattern
        self._is_regex = is_regex
        self._order = order
        self._from_study = from_study
        self._repository = repository
        self._fallback_to_default = fallback_to_default
        # study_ and collection_ are not intended to be provided to __init__
        # except when recreating when using initkwargs
        self._study = study_
        self._collection = collection_

    def __eq__(self, other):
        return (self.from_study == other.from_study and
                self.pattern == other.pattern and
                self.is_regex == other.is_regex and
                self.order == other.order and
                self._repository == other._repository and
                self._fallback_to_default == other._fallback_to_default)

    def __hash__(self):
        return (hash(self.from_study) ^
                hash(self.pattern) ^
                hash(self.is_regex) ^
                hash(self.order) ^
                hash(self._repository) ^
                hash(self._fallback_to_default))

    def initkwargs(self):
        dct = {}
        dct['from_study'] = self.from_study
        dct['pattern'] = self.pattern
        dct['order'] = self.order
        dct['is_regex'] = self.is_regex
        dct['study_'] = self._study
        dct['collection_'] = self._collection
        dct['fallback_to_default'] = self._fallback_to_default
        return dct

    @property
    def pattern(self):
        return self._pattern

    @property
    def derived(self):
        return self._from_study is not None

    @property
    def from_study(self):
        return self._from_study

    @property
    def fallback_to_default(self):
        return self._fallback_to_default

    @property
    def repository(self):
        if self._repository is None:
            if self._study is None:
                raise ArcanaUsageError(
                    "Cannot access repository of {} as it wasn't explicitly "
                    "provided and Selector hasn't been bound to a study"
                    .format(self))
            repo = self._study.repository
        else:
            repo = self._repository
        return repo

    @property
    def collection(self):
        if self._collection is None:
            raise ArcanaUsageError(
                "{} has not been bound to a study".format(self))
        return self._collection

    @property
    def is_regex(self):
        return self._is_regex

    @property
    def order(self):
        return self._order

    def bind(self, study, **kwargs):
        if self._study == study:
            bound = self
        else:
            # Use the default study repository if not explicitly
            # provided to match
            if self.fallback_to_default:
                spec = study.data_spec(self.name)
                if not spec.derived and spec.default is None:
                    raise ArcanaUsageError(
                        "Cannot fallback to default for '{}' spec in {} as it "
                        " is not derived and doesn't have a default"
                        .format(self.name, study))
            if self._repository is None:
                tree = study.tree
            else:
                tree = self._repository.cached_tree(
                    subject_ids=study.subject_ids,
                    visit_ids=study.visit_ids)
            # Match against tree
            bound = copy(self)
            bound._study = study
            bound._collection = self.match(tree, **kwargs)
        return bound

    @property
    def prefixed_name(self):
        return self.name

    def basename(self, **kwargs):
        if not self.is_regex:
            basename = self.pattern
        else:
            basename = self.match(**kwargs).name
        return basename

    def match(self, tree, **kwargs):
        # Run the match against the tree
        if self.frequency == 'per_session':
            nodes = chain(*(s.sessions for s in tree.subjects))
        elif self.frequency == 'per_subject':
            nodes = tree.subjects
        elif self.frequency == 'per_visit':
            nodes = tree.visits
        elif self.frequency == 'per_study':
            nodes = [tree]
        else:
            assert False, "Unrecognised frequency '{}'".format(
                self.frequency)
        return self.CollectionClass(
            self.name, (self._match_node(n, **kwargs)
                        for n in nodes),
            frequency=self.frequency,
            **self._specific_collection_kwargs)

    def _match_node(self, node, **kwargs):
        # Get names matching pattern
        matches = self._filtered_matches(node, **kwargs)
        # Filter matches by study name
        matches = [d for d in matches
                   if d.from_study == self.from_study]
        # Select the fileset from the matches
        if self.order is not None:
            try:
                match = matches[self.order]
            except IndexError:
                raise ArcanaFilesetSelectorError(
                    "Did not find {} filesets names matching pattern {}"
                    " (found {}) in {}".format(self.order, self.pattern,
                                               len(matches), node))
        elif len(matches) == 1:
            match = matches[0]
        elif matches:
            raise ArcanaFilesetSelectorError(
                "Found multiple matches for {} pattern in {} ({})"
                .format(self.pattern, node,
                        ', '.join(str(m) for m in matches)))
        else:
            raise ArcanaFilesetSelectorError(
                "Did not find any matches for {} pattern in {} "
                "(found {})"
                .format(self.pattern, node,
                        ', '.join(str(d) for d in node.filesets)))
        return match


class FilesetSelector(BaseSelector, BaseFileset):
    """
    A pattern that describes a single fileset (typically acquired
    rather than generated but not necessarily) within each session.

    Parameters
    ----------
    name : str
        The name of the fileset, typically left None and set in Study
    format : FileFormat
        The file format used to store the fileset. Can be one of the
        recognised formats
    pattern : str
        A regex pattern to match the fileset names with. Must match
        one and only one fileset per <frequency>. If None, the name
        is used instead.
    is_regex : bool
        Flags whether the pattern is a regular expression or not
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    id : int | None
        To be used to distinguish multiple filesets that match the
        pattern in the same session. The ID of the fileset within the
        session.
    order : int | None
        To be used to distinguish multiple filesets that match the
        pattern in the same session. The order of the fileset within the
        session. Based on the scan ID but is more robust to small
        changes to the IDs within the session if for example there are
        two scans of the same type taken before and after a task.
    dicom_tags : dct(str | str)
        To be used to distinguish multiple filesets that match the
        pattern in the same session. The provided DICOM values dicom
        header values must match exactly.
    from_study : str
        The name of the study that generated the derived fileset to match.
        Is used to determine the location of the filesets in the
        repository as the derived filesets and fields are grouped by
        the name of the study that generated them.
    fallback_to_default : bool
        If the there is no fileset/field matching the selection for a node
        and corresponding data spec has a default or is a derived spec,
        then fallback to the default or generate the derivative.
    repository : BaseRepository | None
        The repository to draw the matches from, if not the main repository
        that is used to store the products of the current study.
    """

    is_spec = False
    CollectionClass = FilesetCollection

    def __init__(self, name, format, pattern=None, # @ReservedAssignment @IgnorePep8
                 frequency='per_session', id=None,  # @ReservedAssignment
                 order=None, dicom_tags=None, is_regex=False, from_study=None,
                 fallback_to_default=False, repository=None, study_=None,
                 collection_=None):
        if pattern is None and id is None:
            raise ArcanaUsageError(
                "Either 'pattern' or 'id' need to be provided to "
                "FilesetSelector constructor")
        BaseFileset.__init__(self, name, format, frequency)
        BaseSelector.__init__(self, pattern, is_regex, order,
                              from_study, fallback_to_default, repository,
                              study_, collection_)
        if dicom_tags is not None and format.name != 'dicom':
            raise ArcanaUsageError(
                "Cannot use 'dicom_tags' kwarg with non-DICOM "
                "format ({})".format(format))
        self._dicom_tags = dicom_tags
        if order is not None and id is not None:
            raise ArcanaUsageError(
                "Cannot provide both 'order' and 'id' to a fileset"
                "match")
        self._id = str(id) if id is not None else id

    def __eq__(self, other):
        return (BaseFileset.__eq__(self, other) and
                BaseSelector.__eq__(self, other) and
                self.dicom_tags == other.dicom_tags and
                self.id == other.id)

    def __hash__(self):
        return (BaseFileset.__hash__(self) ^
                BaseSelector.__hash__(self) ^
                hash(self.dicom_tags) ^
                hash(self.id))

    def initkwargs(self):
        dct = BaseFileset.initkwargs(self)
        dct.update(BaseSelector.initkwargs(self))
        dct['dicom_tags'] = self.dicom_tags
        dct['id'] = self.id
        return dct

    def __repr__(self):
        return ("{}(name='{}', format={}, frequency={}, pattern={}, "
                "is_regex={}, order={}, id={}, dicom_tags={}, "
                "from_study={})"
                .format(self.__class__.__name__, self.name, self.format,
                        self.frequency, self._pattern, self.is_regex,
                        self.order, self.id, self.dicom_tags,
                        self._from_study))

    @property
    def id(self):
        return self._id

    def bind(self, study, **kwargs):
        with ExitStack() as stack:
            # If dicom tags are used to match against then a connection
            # to the repository may be required to query them.
            if self.dicom_tags is not None and self._study != study:
                stack.enter_context(study.repository)
            return super(FilesetSelector, self).bind(study, **kwargs)

    @property
    def dicom_tags(self):
        return self._dicom_tags

    def _filtered_matches(self, node):
        if self.pattern is not None:
            if self.is_regex:
                pattern_re = re.compile(self.pattern)
                matches = [d for d in node.filesets
                           if pattern_re.match(d.name)]
            else:
                matches = [d for d in node.filesets
                           if d.name == self.pattern]
        else:
            matches = list(node.filesets)
        if not matches:
            raise ArcanaFilesetSelectorError(
                "No fileset names in {}:{} match '{}' pattern, found: {}"
                .format(node.subject_id, node.visit_id, self.pattern,
                        ', '.join(d.name for d in node.filesets)))
        if self.id is not None:
            filtered = [d for d in matches if d.id == self.id]
            if not filtered:
                raise ArcanaFilesetSelectorError(
                    "Did not find filesets names matching pattern {} "
                    "with an id of {} (found {}) in {}".format(
                        self.pattern, self.id,
                        ', '.join(str(m) for m in matches), node))
            matches = filtered
        # Filter matches by dicom tags
        if self.dicom_tags is not None:
            filtered = []
            for fileset in matches:
                values = fileset.dicom_values(
                    list(self.dicom_tags.keys()))
                if self.dicom_tags == values:
                    filtered.append(fileset)
            if not filtered:
                raise ArcanaFilesetSelectorError(
                    "Did not find filesets names matching pattern {}"
                    "that matched DICOM tags {} (found {}) in {}"
                    .format(self.pattern, self.dicom_tags,
                            ', '.join(str(m) for m in matches), node))
            matches = filtered
        return matches

    @property
    def _specific_collection_kwargs(self):
        return {'format': self.format}


class FieldSelector(BaseSelector, BaseField):
    """
    A pattern that matches a single field (typically acquired rather than
    generated but not necessarily) in each session.

    Parameters
    ----------
    name : str
        The name of the fileset
    dtype : type
        The datatype of the value. Can be one of (float, int, str)
    pattern : str
        A regex pattern to match the field names with. Must match
        one and only one fileset per <frequency>. If None, the name
        is used instead.
    is_regex : bool
        Flags whether the pattern is a regular expression or not
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    order : int | None
        To be used to distinguish multiple filesets that match the
        pattern in the same session. The order of the fileset within the
        session. Based on the scan ID but is more robust to small
        changes to the IDs within the session if for example there are
        two scans of the same type taken before and after a task.
    from_study : str
        The name of the study that generated the derived field to match.
        Is used to determine the location of the fields in the
        repository as the derived filesets and fields are grouped by
        the name of the study that generated them.
    fallback_to_default : bool
        If the there is no fileset/field matching the selection for a node
        and corresponding data spec has a default or is a derived spec,
        then fallback to the default or generate the derivative.
    repository : BaseRepository | None
        The repository to draw the matches from, if not the main repository
        that is used to store the products of the current study.
    """

    is_spec = False
    CollectionClass = FieldCollection

    def __init__(self, name, dtype, pattern, frequency='per_session',
                 order=None, is_regex=False, from_study=None,
                 fallback_to_default=False, repository=None, study_=None,
                 collection_=None):
        BaseField.__init__(self, name, dtype, frequency)
        BaseSelector.__init__(self, pattern, is_regex, order,
                              from_study, fallback_to_default, repository,
                              study_, collection_)

    def __eq__(self, other):
        return (BaseField.__eq__(self, other) and
                BaseSelector.__eq__(self, other))

    def __hash__(self):
        return (BaseField.__hash__(self) ^ BaseSelector.__hash__(self))

    def initkwargs(self):
        dct = BaseField.initkwargs(self)
        dct.update(BaseSelector.initkwargs(self))
        return dct

    def _filtered_matches(self, node):
        if self.is_regex:
            pattern_re = re.compile(self.pattern)
            matches = [f for f in node.fields
                       if pattern_re.match(f.name)]
        else:
            matches = [f for f in node.fields
                       if f.name == self.pattern]
        if self.from_study is not None:
            matches = [f for f in matches
                       if f.from_study == self.from_study]
        if not matches:
            raise ArcanaFilesetSelectorError(
                "No field names in {} match '{}' pattern"
                .format(node, self.pattern))
        return matches

    def __repr__(self):
        return ("{}(name='{}', dtype={}, frequency={}, pattern={}, "
                "is_regex={}, order={}, from_study={})"
                .format(self.__class__.__name__, self.name, self.dtype,
                        self.frequency, self._pattern, self.is_regex,
                        self.order, self._from_study))

    @property
    def _specific_collection_kwargs(self):
        return {'dtype': self.dtype}



from __future__ import absolute_import
from builtins import str  # @UnusedImport
import os.path as op
from collections import defaultdict
from itertools import chain
from arcana.utils import makedirs
from .directory import DirectoryRepository
import logging
from bids import grabbids as gb
from .tree import Tree, Subject, Session, Visit
from arcana.data import Fileset
from arcana.exceptions import ArcanaNameError

logger = logging.getLogger('arcana')


class BidsRepository(DirectoryRepository):
    """
    An 'Repository' class for directories on the local file system organised
    into sub-directories by subject and then visit.

    Parameters
    ----------
    base_dir : str (path)
        Path to local directory containing data
    """

    type = 'bids'
    DERIVATIVES_SUB_PATH = op.join('derivatives')

    def __init__(self, root_dir):
        self._root_dir = root_dir
        derivatives_path = op.join(root_dir,
                                        self.DERIVATIVES_SUB_PATH)
        makedirs(derivatives_path, exist_ok=True)
        DirectoryRepository.__init__(derivatives_path)
        self._layout = gb.BIDSLayout(self.base_dir)

    @property
    def root_dir(self):
        return self._root_dir

    @property
    def layout(self):
        return self._layout

    def __repr__(self):
        return "BidsRepository(root_dir='{}')".format(self.root_dir)

    def __eq__(self, other):
        try:
            return self.root_dir == other.root_dir
        except AttributeError:
            return False

    def get_fileset(self, fileset):
        """
        Set the path of the fileset from the repository
        """
        raise NotImplementedError

    def tree(self, subject_ids=None, visit_ids=None):
        """
        Return subject and session information for a project in the local
        repository

        Parameters
        ----------
        subject_ids : list(str)
            List of subject IDs with which to filter the tree with. If None all
            are returned
        visit_ids : list(str)
            List of visit IDs with which to filter the tree with. If None all
            are returned

        Returns
        -------
        project : arcana.repository.Tree
            A hierarchical tree of subject, session and fileset information for
            the repository
        """
        bids_filesets = defaultdict(lambda: defaultdict(dict))
        derived_tree = super(BidsRepository, self).tree(
            subject_ids=None, visit_ids=None)
        for bids_obj in self.layout.get(return_type='object'):
            subj_id = bids_obj.entities['subject']
            if subject_ids is not None and subj_id not in subject_ids:
                continue
            visit_id = bids_obj.entities['session']
            if visit_ids is not None and visit_id not in visit_ids:
                continue
            bids_filesets[subj_id][visit_id] = Fileset.from_path(
                bids_obj.path, frequency='per_session',
                subject_id=subj_id, visit_id=visit_id, repository=self,
                bids_attrs=bids_obj)
        # Need to pull out all filesets and fields
        all_sessions = defaultdict(dict)
        all_visit_ids = set()
        for subj_id, visits in bids_filesets.items():
            for visit_id, filesets in visits.items():
                session = Session(
                    subject_id=subj_id, visit_id=visit_id,
                    filesets=filesets)
                try:
                    session.derived = derived_tree.subject(
                        subj_id).visit(visit_id)
                except ArcanaNameError:
                    pass  # No matching derived session
                all_sessions[subj_id][visit_id] = session
                all_visit_ids.add(visit_id)

        subjects = []
        for subj_id, subj_sessions in list(all_sessions.items()):
            try:
                derived_subject = derived_tree.subject(subj_id)
            except ArcanaNameError:
                filesets = []
                fields = []
            else:
                filesets = derived_subject.filesets
                fields = derived_subject.fields
            subjects.append(Subject(
                subj_id, sorted(subj_sessions.values()),
                filesets, fields))
        visits = []
        for visit_id in all_visit_ids:
            try:
                derived_visit = derived_tree.visit(subj_id)
            except ArcanaNameError:
                filesets = []
                fields = []
            else:
                filesets = derived_visit.filesets
                fields = derived_visit.fields
            visit_sessions = list(chain(
                sess[visit_id] for sess in list(all_sessions.values())))
            visits.append(
                Visit(visit_id, sorted(visit_sessions),
                      filesets, fields))
        return Tree(sorted(subjects), sorted(visits),
                       derived_tree.filesets, derived_tree.fields)



from arcana.exceptions import ArcanaFilesetSelectorError, ArcanaUsageError
from .selector import FilesetSelector


class BidsAttrs(object):

    def __init__(self, type=None, modality=None, run=None, metadata=None,  # @ReservedAssignment @IgnorePep8
                 field_maps=None, bvec=None, bval=None, desc=None):
        self._type = type
        self._modality = modality
        self._run = run
        self._metadata = metadata
        self._bval = bval
        self._bvec = bvec
        self._field_maps = field_maps
        self._desc = desc

    @property
    def type(self):
        return self._type

    @property
    def mode(self):
        return self._mode

    @property
    def run(self):
        return self._run

    @property
    def metadata(self):
        return self._metadata

    @property
    def bval(self):
        return self._bval

    @property
    def bvec(self):
        return self._bvec

    @property
    def field_maps(self):
        return self._field_maps

    @property
    def desc(self):
        return self._desc


class BidsMatch(FilesetSelector):
    """
    A match object for matching filesets from their 'bids_attr'
    attribute

    Parameters
    ----------
    name : str
        Name of the fileset
    type : str
        Type of the fileset
    modality : str
        Modality of the filesets
    format : FileFormat
        The file format of the fileset to match
    run : int
        Run number of the fileset
    """

    def __init__(self, name, type, modality, format, run=None):  # @ReservedAssignment @IgnorePep8
        FilesetSelector.__init__(
            self, name, format, pattern=None, frequency='per_session',   # @ReservedAssignment @IgnorePep8
            id=None, order=run, dicom_tags=None, is_regex=False,
            from_study=None)
        self._type = type
        self._modality = modality
        self._run = run

    @property
    def type(self):
        return self._type

    @property
    def modality(self):
        return self._modality

    @property
    def run(self):
        return self.order

    def _filtered_matches(self, node):
        matches = [
            d for d in node.filesets
            if (d.bids_attr.entities['type'] == self.type and
                d.bids_attr.entities['modality'] == self.modality)]
        if not matches:
            raise ArcanaFilesetSelectorError(
                "No BIDS filesets for subject={}, visit={} match "
                "modality '{}' and type '{}' found:\n{}"
                .format(node.subject_id, node.visit_id, self.modality,
                        self.type, '\n'.join(
                            sorted(d.name for d in node.filesets))))
        return matches

    def __eq__(self, other):
        return (FilesetSelector.__eq__(self, other) and
                self.type == other.type and
                self.modality == other.modality and
                self.run == other.run)

    def __hash__(self):
        return (FilesetSelector.__hash__(self) ^
                hash(self.type) ^
                hash(self.modality) ^
                hash(self.run))

    def initkwargs(self):
        dct = FilesetSelector.initkwargs(self)
        dct['type'] = self.type
        dct['modality'] = self.modality
        dct['run'] = self.run
        return dct


class BidsAssociatedMatch(FilesetSelector):
    """
    A match object for matching BIDS filesets that are associated with
    another BIDS filesets (e.g. field-maps, bvecs, bvals)

    Parameters
    ----------
    name : str
        Name of the associated fileset
    primary_match : BidsMatch
        The primary fileset which the fileset to match is associated with
    associated : str
        The name of the association between the fileset to match and the
        primary fileset
    fieldmap_type : str
        Key of the return fieldmap dictionary (if association=='fieldmap'
    order : int
        Order of the fieldmap dictionary that you want to match
    """

    VALID_ASSOCIATIONS = ('fieldmap', 'bvec', 'bval')

    def __init__(self, name, primary_match, format, association,  # @ReservedAssignment @IgnorePep8
                 fieldmap_type=None, order=0):
        FilesetSelector.__init__(
            self, name, format, pattern=None, frequency='per_session',   # @ReservedAssignment @IgnorePep8
            id=None, order=order, dicom_tags=None, is_regex=False,
            from_study=None)
        self._primary_match = primary_match
        self._association = association
        if fieldmap_type is not None and association != 'fieldmap':
            raise ArcanaUsageError(
                "'fieldmap_type' (provided to '{}' match) "
                "is only valid for 'fieldmap' "
                "associations (not '{}')".format(name, association))
        self._fieldmap_type = fieldmap_type

    def __repr__(self):
        return ("{}(name={}, primary_match={}, format={}, association={}, "
                "fieldmap_type\{}, order={})".format(
                    self.name, self.primary_match, self.format,
                    self.association, self.fieldmap_type,
                    self.order))

    @property
    def primary_match(self):
        return self._primary_match

    @property
    def association(self):
        return self._association

    @property
    def fieldmap_type(self):
        return self._fieldmap_type

    def _bind_node(self, node):
        primary_match = self._primary_match._bind_node(node)
        layout = self.study.repository.layout
        if self._association == 'fieldmap':
            matches = layout.get_fieldmap(primary_match.path,
                                          return_list=True)
            try:
                match = matches[0]
            except IndexError:
                raise ArcanaFilesetSelectorError(
                    "Provided order to associated BIDS fileset match "
                    "{} is out of range")
        elif self._association == 'bvec':
            match = layout.get_bvec(primary_match.path)
        elif self._association == 'bval':
            match = layout.get_bval(primary_match.path)
        return match
        return matches

    def __eq__(self, other):
        return (FilesetSelector.__eq__(self, other) and
                self.primary_match == other.primary_match and
                self.association == other.association and
                self.fieldmap_type == other.fieldmap_type)

    def __hash__(self):
        return (FilesetSelector.__hash__(self) ^
                hash(self.primary_match) ^
                hash(self.association) ^
                hash(self.fieldmap_type))

    def initkwargs(self):
        dct = FilesetSelector.initkwargs(self)
        dct['primary_match'] = self.primary_match
        dct['association'] = self.association
        dct['fieldmap_type'] = self.fieldmap_type
        return dct



from copy import copy
from arcana.utils.interfaces import (
    ZipDir, UnzipDir, TarGzDir, UnTarGzDir)
from .base import FileFormat, Converter


class UnzipConverter(Converter):

    interface = UnzipDir()
    mem_gb = 12
    input = 'zipped'
    output = 'unzipped'


class ZipConverter(Converter):

    interface = ZipDir()
    mem_gb = 12
    input = 'dirname'
    output = 'zipped'


class TarGzConverter(Converter):

    interface = TarGzDir()
    mem_gb = 12
    input = 'dirname'
    output = 'zipped'


class UnTarGzConverter(Converter):

    interface = UnTarGzDir()
    mem_gb = 12
    input = 'gzipped'
    output = 'gunzipped'


# General formats
directory_format = FileFormat(name='directory', extension=None,
                              directory=True,
                              converters={'zip': UnzipConverter,
                                          'targz': UnTarGzConverter})
text_format = FileFormat(name='text', extension='.txt')


# Compressed formats
zip_format = FileFormat(name='zip', extension='.zip',
                        converters={'directory': ZipConverter})
targz_format = FileFormat(name='targz', extension='.tar.gz',
                          converters={'direcctory': TarGzConverter})

# Register all data formats in module
for file_format in copy(globals()).values():
    if isinstance(file_format, FileFormat):
        FileFormat.register(file_format)



from past.builtins import basestring
from builtins import object
from operator import attrgetter
from copy import copy, deepcopy
from arcana.exceptions import (
    ArcanaError, ArcanaUsageError,
    ArcanaOutputNotProducedException,
    ArcanaMissingDataException, ArcanaNameError)
from .base import BaseFileset, BaseField
from .item import Fileset, Field
from .collection import FilesetCollection, FieldCollection


class BaseAcquiredSpec(object):

    derived = False

    def __init__(self, name, desc=None, optional=False, default=None):
        if optional and default is not None:
            raise ArcanaUsageError(
                "'optional' doesn't make sense for specs ('{}') with default "
                "values"
                .format(name))
        self._desc = desc
        self._study = None
        self._optional = optional
        # Set the name of the default collection-like object so it matches
        # the name of the spec
        if default is not None:
            if default.frequency != self.frequency:
                raise ArcanaUsageError(
                    "Frequency of default collection-like object passed to "
                    "'{}' spec ('{}'), does not match spec ('{}')".format(
                        name, default.frequency, self.frequency))
            default = deepcopy(default)
        self._default = default

    def __eq__(self, other):
        return (self.desc == other.desc and
                self.optional == other.optional and
                self.default == other.default)

    def __hash__(self):
        return (hash(self.desc) ^ hash(self.optional) ^ hash(self.default))

    def initkwargs(self):
        dct = {}
        dct['desc'] = self.desc
        dct['optional'] = self.optional
        dct['default'] = deepcopy(self.default)
        return dct

    def find_mismatch(self, other, indent=''):
        mismatch = ''
        sub_indent = indent + '  '
        if self.optional != other.optional:
            mismatch += ('\n{}optional: self={} v other={}'
                         .format(sub_indent, self.optional,
                                 other.optional))
        if self.default != other.default:
            mismatch += ('\n{}default: self={} v other={}'
                         .format(sub_indent, self.default,
                                 other.default))
        if self.desc != other.desc:
            mismatch += ('\n{}desc: self={} v other={}'
                         .format(sub_indent, self.desc, other.desc))
        return mismatch

    def bind(self, study):
        """
        Returns a copy of the AcquiredSpec bound to the given study

        Parameters
        ----------
        study : Study
            A study to bind the fileset spec to (should happen in the
            study __init__)
        """
        if self.default is None:
            raise ArcanaError(
                "Attempted to bind '{}' to {} but only acquired specs with "
                "a default value should be bound to studies{})".format(
                    self.name, study))
        if self._study is not None:
            # This avoids rebinding specs to sub-studies that have already
            # been bound to the multi-study
            bound = self
        else:
            bound = copy(self)
            bound._study = study
            bound._default = bound.default.bind(study)
        return bound

    @property
    def optional(self):
        return self._optional

    @property
    def study(self):
        if self._study is None:
            raise ArcanaError(
                "{} is not bound to a study".format(self))
        return self._study

    @property
    def default(self):
        # Ensure the default has the same name as the spec to it can be
        # accessed properly by the source node
        if self._default is not None:
            # FIXME: Should use setter but throwing an error for some reason
            self._default.name = self.name
        return self._default

    @property
    def desc(self):
        return self._desc

    @property
    def collection(self):
        if self._study is None:
            raise ArcanaUsageError(
                "{} needs to be bound to a study before accessing "
                "the corresponding collection".format(self))
        return self.default.collection


class BaseSpec(object):

    derived = True

    def __init__(self, name, pipeline_getter, desc=None):
        if pipeline_getter is not None:
            if not isinstance(pipeline_getter, basestring):
                raise ArcanaUsageError(
                    "Pipeline name for {} '{}' is not a string "
                    "'{}'".format(name, pipeline_getter))
        self._pipeline_getter = pipeline_getter
        self._desc = desc
        self._study = None
        self._collection = None

    def __eq__(self, other):
        return (self.pipeline_getter == other.pipeline_getter and
                self.desc == other.desc)

    def __hash__(self):
        return (hash(self.pipeline_getter) ^ hash(self.desc))

    def initkwargs(self):
        dct = {}
        dct['pipeline_getter'] = self.pipeline_getter
        dct['desc'] = self.desc
        return dct

    def find_mismatch(self, other, indent=''):
        mismatch = ''
        sub_indent = indent + '  '
        if self.pipeline_getter != other.pipeline_getter:
            mismatch += ('\n{}pipeline_getter: self={} v other={}'
                         .format(sub_indent, self.pipeline_getter,
                                 other.pipeline_getter))
        if self.desc != other.desc:
            mismatch += ('\n{}desc: self={} v other={}'
                         .format(sub_indent, self.desc, other.desc))
        return mismatch

    def bind(self, study):
        """
        Returns a copy of the Spec bound to the given study

        Parameters
        ----------
        study : Study
            A study to bind the fileset spec to (should happen in the
            study __init__)
        """
        if self._study is not None:
            # Avoid rebinding specs in sub-studies that have already
            # been bound to MultiStudy
            bound = self
        else:
            bound = copy(self)
            bound._study = study
            if not hasattr(study, self.pipeline_getter):
                raise ArcanaError(
                    "{} does not have a method named '{}' required to "
                    "derive {}".format(study, self.pipeline_getter,
                                       self))
            bound._bind_tree(study.tree)
        return bound

    @property
    def collection(self):
        if self._collection is None:
            raise ArcanaUsageError(
                "{} needs to be bound to a study before accessing "
                "the corresponding collection".format(self))
        return self._collection

    def _bind_tree(self, tree, **kwargs):
        # Run the match against the tree
        if self.frequency == 'per_session':
            nodes = []
            for subject in tree.subjects:
                for sess in subject.sessions:
                    nodes.append(sess)
        elif self.frequency == 'per_subject':
            nodes = tree.subjects
        elif self.frequency == 'per_visit':
            nodes = tree.visits
        elif self.frequency == 'per_study':
            nodes = [tree]
        else:
            assert False, "Unrecognised frequency '{}'".format(
                self.frequency)
        self._collection = self.CollectionClass(
            self.name, (self._bind_node(n, **kwargs) for n in nodes),
            frequency=self.frequency,
            **self._specific_collection_kwargs)

    @property
    def derivable(self):
        """
        Whether the spec (only valid for derived specs) can be derived
        given the inputs and switches provided to the study
        """
        try:
            # Just need to iterate all study inputs and catch relevant
            # exceptions
            list(self.pipeline.study_inputs)
        except (ArcanaOutputNotProducedException,
                ArcanaMissingDataException):
            return False
        return True

    @property
    def pipeline_getter(self):
        return self._pipeline_getter

    @property
    def pipeline(self):
        return self.study.pipeline(self.pipeline_getter, [self.name])

    @property
    def study(self):
        if self._study is None:
            raise ArcanaError(
                "{} is not bound to a study".format(self))
        return self._study

    @property
    def desc(self):
        return self._desc

    def _tree_node(self, subject_id=None, visit_id=None):
        if self.frequency == 'per_session':
            node = self.study.tree.subject(subject_id).session(visit_id)
        elif self.frequency == 'per_subject':
            node = self.study.tree.subject(subject_id)
        elif self.frequency == 'per_visit':
            node = self.study.tree.visit(visit_id)
        elif self.frequency == 'per_study':
            node = self.study.tree
        else:
            assert False
        return node


class AcquiredFilesetSpec(BaseFileset, BaseAcquiredSpec):
    """
    A specification for an "acquired" fileset (e.g from the scanner or
    standard atlas)

    Parameters
    ----------
    name : str
        The name of the fileset
    valid_formats : FileFormat | list[FileFormat]
        The acceptable file formats for input filesets to match this spec
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    desc : str
        Description of what the field represents
    optional : bool
        Whether the specification is optional or not. Only valid for
        "acquired" fileset specs.
    default : FilesetCollection | object
        The default value to be passed as an input to this spec if none are
        provided. Can either be an explicit FilesetCollection or any object
        with a 'collection' property that will return a default collection.
        This object should also implement a 'bind(self, study)' method to
        allow the study to be bound to it.
    """

    is_spec = True
    CollectionClass = FilesetCollection

    def __init__(self, name, valid_formats, frequency='per_session', # @ReservedAssignment @IgnorePep8
                 desc=None, optional=False, default=None):
        # Ensure allowed formats is a list
        try:
            valid_formats = sorted(valid_formats, key=attrgetter('name'))
        except TypeError:
            valid_formats = [valid_formats]
        else:
            if not valid_formats:
                raise ArcanaError(
                    "'{}' spec doesn't have any allowed formats".format(name))
        self._valid_formats = valid_formats
        BaseFileset.__init__(self, name, valid_formats[0], frequency)
        BaseAcquiredSpec.__init__(self, name, desc, optional=optional,
                                  default=default)

    @property
    def valid_formats(self):
        return iter(self._valid_formats)

    def __eq__(self, other):
        return (BaseFileset.__eq__(self, other) and
                BaseAcquiredSpec.__eq__(self, other) and
                self._valid_formats == other._valid_formats)

    def __hash__(self):
        return (BaseFileset.__hash__(self) ^
                BaseAcquiredSpec.__hash__(self) and
                hash(self.valid_formats))

    def initkwargs(self):
        dct = BaseFileset.initkwargs(self)
        del dct['format']
        dct.update(BaseAcquiredSpec.initkwargs(self))
        dct['valid_formats'] = self._valid_formats
        return dct

    def __repr__(self):
        return ("{}(name='{}', valid_formats={}, frequency={}, default={}, "
                "optional={})"
                .format(type(self).__name__, self.name,
                        list(f.name for f in self.valid_formats),
                        self.frequency, self.default, self.optional))

    def find_mismatch(self, other, indent=''):
        sub_indent = indent + '  '
        mismatch = BaseFileset.find_mismatch(self, other, indent)
        mismatch += BaseAcquiredSpec.find_mismatch(self, other, indent)
        if self.valid_formats != other.valid_formats:
            mismatch += ('\n{}pipeline: self={} v other={}'
                         .format(sub_indent, list(self.valid_formats),
                                 list(other.valid_formats)))
        return mismatch


class FilesetSpec(BaseFileset, BaseSpec):
    """
    A specification for a fileset within a study to be derived from a
    processing pipeline.

    Parameters
    ----------
    name : str
        The name of the fileset
    format : FileFormat
        The file format used to store the fileset. Can be one of the
        recognised formats
    pipeline_getter : str
        Name of the method in the study that constructs a pipeline to derive
        the fileset
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    desc : str
        Description of what the field represents
    """

    is_spec = True
    CollectionClass = FilesetCollection

    def __init__(self, name, format, pipeline_getter, frequency='per_session',  # @ReservedAssignment @IgnorePep8 
                 desc=None):
        BaseFileset.__init__(self, name, format, frequency)
        BaseSpec.__init__(self, name, pipeline_getter, desc)

    def __eq__(self, other):
        return (BaseFileset.__eq__(self, other) and
                BaseSpec.__eq__(self, other))

    def __hash__(self):
        return (BaseFileset.__hash__(self) ^
                BaseSpec.__hash__(self))

    def initkwargs(self):
        dct = BaseFileset.initkwargs(self)
        dct.update(BaseSpec.initkwargs(self))
        return dct

    def __repr__(self):
        return ("FilesetSpec(name='{}', format={}, pipeline_getter={}, "
                "frequency={})".format(
                    self.name, self.format, self.pipeline_getter,
                    self.frequency))

    def find_mismatch(self, other, indent=''):
        mismatch = BaseFileset.find_mismatch(self, other, indent)
        mismatch += BaseSpec.find_mismatch(self, other, indent)
        return mismatch

    def _bind_node(self, node, **kwargs):
        try:
            fileset = node.fileset(self.name, from_study=self.study.name)
        except ArcanaNameError:
            # For filesets that can be generated by the analysis
            fileset = Fileset(self.name, format=self.format,
                              frequency=self.frequency, path=None,
                              subject_id=node.subject_id,
                              visit_id=node.visit_id,
                              repository=self.study.repository,
                              from_study=self.study.name,
                              exists=False,
                              **kwargs)
        return fileset

    @property
    def _specific_collection_kwargs(self):
        return {'format': self.format}


class AcquiredFieldSpec(BaseField, BaseAcquiredSpec):
    """
    An abstract base class representing an acquired field

    Parameters
    ----------
    name : str
        The name of the fileset
    dtype : type
        The datatype of the value. Can be one of (float, int, str)
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    desc : str
        Description of what the field represents
    optional : bool
        Whether the specification is optional or not. Only valid for
        "acquired" fileset specs.
    default : FieldCollection | callable
        The default value to be passed as an input to this spec if none are
        provided. Can either be an explicit FieldCollection or any object
        with a 'collection' property that will return a default collection.
        This object should also implement a 'bind(self, study)' method to
        allow the study to be bound to it.
    """

    is_spec = True
    CollectionClass = FieldCollection

    def __init__(self, name, dtype, frequency='per_session', desc=None,
                 optional=False, default=None, array=False):
        BaseField.__init__(self, name, dtype, frequency, array=array)
        BaseAcquiredSpec.__init__(self, name, desc, optional=optional,
                                  default=default)

    def __eq__(self, other):
        return (BaseField.__eq__(self, other) and
                BaseAcquiredSpec.__eq__(self, other))

    def __hash__(self):
        return (BaseField.__hash__(self) ^ BaseAcquiredSpec.__hash__(self))

    def find_mismatch(self, other, indent=''):
        mismatch = BaseField.find_mismatch(self, other, indent)
        mismatch += BaseAcquiredSpec.find_mismatch(self, other, indent)
        return mismatch

    def __repr__(self):
        return ("{}(name='{}', dtype={}, frequency={}, array={})".format(
            self.__class__.__name__, self.name, self.dtype,
            self.frequency, self.array))

    def initkwargs(self):
        dct = BaseField.initkwargs(self)
        dct.update(BaseAcquiredSpec.initkwargs(self))
        return dct


class FieldSpec(BaseField, BaseSpec):
    """
    An abstract base class representing the specification for a derived
    fileset.

    Parameters
    ----------
    name : str
        The name of the fileset
    dtype : type
        The datatype of the value. Can be one of (float, int, str)
    pipeline_getter : str
        Name of the method that constructs pipelines to derive the field
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    desc : str
        Description of what the field represents
    """

    is_spec = True
    CollectionClass = FieldCollection

    def __init__(self, name, dtype, pipeline_getter=None,
                 frequency='per_session', desc=None, array=False):
        BaseField.__init__(self, name, dtype, frequency, array=array)
        BaseSpec.__init__(self, name, pipeline_getter, desc)

    def __eq__(self, other):
        return (BaseField.__eq__(self, other) and
                BaseSpec.__eq__(self, other))

    def __hash__(self):
        return (BaseField.__hash__(self) ^ BaseSpec.__hash__(self))

    def find_mismatch(self, other, indent=''):
        mismatch = BaseField.find_mismatch(self, other, indent)
        mismatch += BaseSpec.find_mismatch(self, other, indent)
        return mismatch

    def __repr__(self):
        return ("{}(name='{}', dtype={}, pipeline_getter={}, "
                "frequency={}, array={})".format(
                    self.__class__.__name__, self.name, self.dtype,
                    self.pipeline_getter, self.frequency, self.array))

    def initkwargs(self):
        dct = BaseField.initkwargs(self)
        dct.update(BaseSpec.initkwargs(self))
        return dct

    def _bind_node(self, node, **kwargs):
        try:
            field = node.field(self.name, from_study=self.study.name)
        except ArcanaNameError:
            # For fields to be generated by the analysis
            field = Field(self.name, dtype=self.dtype,
                          frequency=self.frequency,
                          subject_id=node.subject_id,
                          visit_id=node.visit_id,
                          repository=self.study.repository,
                          from_study=self.study.name,
                          array=self.array,
                          exists=False, **kwargs)
        return field

    @property
    def _specific_collection_kwargs(self):
        return {'dtype': self.dtype,
                'array': self.array}



from .static import StaticEnvironment
from .modules import ModulesEnvironment
from .requirement import (
    BaseRequirement, CliRequirement, MatlabPackageRequirement,
    PythonPackageRequirement, matlab_req)



import sys
import os.path
from setuptools import setup, find_packages

PACKAGE_NAME = 'arcana'

# Get version from module inside package
sys.path.insert(0, os.path.join(os.path.dirname(__file__),
                                PACKAGE_NAME))
from pkg_info import __version__, install_requires  # @UnresolvedImport @IgnorePep8
sys.path.pop(0)


setup(
    name=PACKAGE_NAME,
    version=__version__,
    author='Tom G. Close',
    author_email='tom.g.close@gmail.com',
    packages=find_packages(),
    url='https://github.com/monashbiomedicalimaging/arcana',
    license='The Apache Software Licence 2.0',
    description=(
        'Abstracted repository-centric analysis framework'),
    long_description=open('README.rst').read(),
    install_requires=install_requires,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Medical Science Apps."],
    keywords='repository analysis')



from .item import Fileset, Field
from .collection import FilesetCollection, FieldCollection
from .spec import (
    FilesetSpec, FieldSpec, AcquiredFilesetSpec, AcquiredFieldSpec, BaseSpec,
    BaseAcquiredSpec)
from .base import BaseField, BaseFileset, BaseData
from .selector import FilesetSelector, FieldSelector, BaseSelector
from .file_format import FileFormat, Converter, IdentityConverter
from .bids import BidsMatch, BidsAssociatedMatch



from builtins import object
from arcana.exceptions import (
    ArcanaUsageError, ArcanaFileFormatClashError, ArcanaNoConverterError,
    ArcanaFileFormatNotRegisteredError, ArcanaMissingDataException)
from nipype.interfaces.utility import IdentityInterface
from arcana.utils import lower, split_extension
import logging


logger = logging.getLogger('arcana')


class FileFormat(object):
    """
    Defines a format for a fileset (e.g. DICOM, NIfTI, Matlab file)

    Parameters
    ----------
    name : str
        A name for the data format
    extension : str
        The extension of the format
    desc : str
        A description of what the format is and ideally a link to its
        documentation
    directory : bool
        Whether the format is a directory or a file
    within_dir_exts : List[str]
        A list of extensions that are found within the top level of
        the directory (for directory formats). Used to identify
        formats from paths.
    converters : Dict[str, Converter]
        A dictionary mapping names of alternative data formats
        to Converter objects that can convert from the alternative
        format to this format.
    alternate_names : List[str]
        A list of alternate names to use to load the file format with
        (when the format is saved by format name, e.g. XNAT, instead
        of a file with an extension)
    array_loader : Function
        A function that takes the fileset path in the given format and
        returns a data array
    """

    # To hold registered data formats
    by_names = {}
    by_exts = {}
    by_within_exts = {}

    def __init__(self, name, extension=None, desc='',
                 directory=False, within_dir_exts=None,
                 converters=None, alternate_names=None,
                 array_loader=None, header_loader=None):
        if not name.islower():
            raise ArcanaUsageError(
                "All data format names must be lower case ('{}')"
                .format(name))
        if extension is None and not directory:
            raise ArcanaUsageError(
                "Extension for '{}' format can only be None if it is a "
                "directory".format(name))
        self._name = name
        self._extension = extension
        self._desc = desc
        self._directory = directory
        if within_dir_exts is not None:
            if not directory:
                raise ArcanaUsageError(
                    "'within_dir_exts' keyword arg is only valid "
                    "for directory data formats, not '{}'".format(name))
            within_dir_exts = frozenset(within_dir_exts)
        self._within_dir_exts = within_dir_exts
        self._converters = converters if converters is not None else {}
        self._alternate_names = (tuple(alternate_names)
                                 if alternate_names is not None else ())
        self._array_loader = array_loader
        self._header_loader = header_loader

    def __eq__(self, other):
        try:
            return (
                self._name == other._name and
                self._extension == other._extension and
                self._desc == other._desc and
                self._directory == other._directory and
                self._within_dir_exts ==
                other._within_dir_exts and
                self.alternate_names == other.alternate_names)
        except AttributeError:
            return False

    def __hash__(self):
        return (
            hash(self._name) ^
            hash(self._extension) ^
            hash(self._desc) ^
            hash(self._directory) ^
            hash(self._within_dir_exts) ^
            hash(self._alternate_names))

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return ("FileFormat(name='{}', extension='{}', directory={}{})"
                .format(self.name, self.extension, self.directory,
                        (', within_dir_extension={}'.format(
                            self.within_dir_exts)
                         if self.directory else '')))

    def __str__(self):
        return self.name

    @property
    def name(self):
        return self._name

    @property
    def extension(self):
        return self._extension

    @property
    def ext(self):
        return self._extension

    @property
    def ext_str(self):
        return self.extension if self.extension is not None else ''

    @property
    def desc(self):
        return self._desc

    @property
    def directory(self):
        return self._directory

    @property
    def alternate_names(self):
        return self._alternate_names

    @property
    def within_dir_exts(self):
        return self._within_dir_exts

    @property
    def xnat_resource_names(self):
        "Lists acceptable XNAT resource names in order of preference"
        return (self.name.upper(),) + self.alternate_names

    def converter_from(self, file_format, **kwargs):
        if file_format == self:
            return IdentityConverter(file_format, self)
        try:
            converter_cls = self._converters[file_format.name]
        except KeyError:
            raise ArcanaNoConverterError(
                "There is no converter to convert {} to {}, available:\n{}"
                .format(file_format, self,
                        '\n'.join(
                            '{} <- {}'.format(k, v)
                            for k, v in self._converters.items())))
        return converter_cls(file_format, self, **kwargs)

    @classmethod
    def register(cls, file_format):
        """
        Registers a data format so they can be recognised by extension
        and by resource type on XNAT

        Parameters
        ----------
        file_format : FileFormat
            The data format to register
        """
        try:
            saved_format = cls.by_names[file_format.name]
            if saved_format != file_format:
                raise ArcanaFileFormatClashError(
                    "Cannot register {} due to name clash with previously "
                    "registered {}".format(file_format, saved_format))
        except KeyError:
            if file_format.directory and file_format.extension is None:
                if file_format.within_dir_exts in cls.by_within_exts:
                    raise ArcanaFileFormatClashError(
                        "Cannot register {} due to within-directory "
                        "extension clash with previously registered {}"
                        .format(file_format,
                                cls.by_within_exts[
                                    file_format.within_dir_exts]))
            else:
                if file_format.extension in cls.by_exts:
                    raise ArcanaFileFormatClashError(
                        "Cannot register {} due to extension clash with "
                        "previously registered {}".format(
                            file_format,
                            cls.by_exts[file_format.ext]))
            for alt_name in file_format.alternate_names:
                if alt_name in cls.by_names:
                    raise ArcanaFileFormatClashError(
                        "Cannot register {} due to alternate name clash"
                        "('{}') with previously registered {}".format(
                            file_format, alt_name,
                            cls.by_names[alt_name]))
            cls.by_names[file_format.name] = file_format
            for alt_name in file_format.alternate_names:
                cls.by_names[alt_name] = file_format
            if file_format.ext is not None:
                cls.by_exts[file_format.ext] = file_format
            if file_format.within_dir_exts is not None:
                cls.by_within_exts[
                    file_format.within_dir_exts] = file_format

    @classmethod
    def by_name(cls, name):
        try:
            return cls.by_names[name.lower()]
        except KeyError:
            raise ArcanaFileFormatNotRegisteredError(
                "No data format named '{}' has been registered"
                .format(name,
                        ', '.format(repr(f)
                                    for f in list(cls.by_names.values()))))

    @classmethod
    def by_ext(cls, ext):
        try:
            return cls.by_exts[ext]
        except KeyError:
            raise ArcanaFileFormatNotRegisteredError(
                "No data format with extension '{}' has been registered"
                .format(
                    ext, ', '.format(repr(f)
                                     for f in list(cls.by_exts.values()))))

    @classmethod
    def by_within_dir_exts(cls, within_exts):
        try:
            return cls.by_within_exts[within_exts]
        except KeyError:
            raise ArcanaFileFormatNotRegisteredError(
                "No data format with within-directory extension '{}' "
                "has been registered ({})".format(
                    within_exts,
                    ', '.format(repr(f)
                                for f in list(cls.by_within_exts.values()))))

    def get_array(self, path):
        """
        Returns array data associated with the given path for the
        file format
        """
        if self._array_loader is None:
            raise ArcanaUsageError(
                "Cannot load array data for '{}' as '{}' file format "
                "doesn't have a registered array loader".format(
                    path, self.name))
        return self._array_loader(path)

    def get_header(self, path):
        """
        Returns header data associated with the given path for the
        file format
        """
        if self._header_loader is None:
            raise ArcanaUsageError(
                "Cannot load header for '{}' as '{}' file format "
                "doesn't have a registered header loader".format(
                    path, self.name))
        return self._header_loader(path)

    def primary_file(self, file_names):
        """
        Selects the primary file from a list of filenames, as opposed to
        side-cars and headers

        Parameters
        ----------
        file_names : list[str]
            The list of filenames to select from

        Returns
        -------
        fname : str
            The name of the primary file
        """

        match_fnames = [
            f for f in file_names
            if lower(split_extension(f)[-1]) == lower(self.extension)]
        if len(match_fnames) == 1:
            fname = match_fnames[0]
        else:
            raise ArcanaMissingDataException(
                "Did not find single file with extension '{}' "
                "(found '{}')"
                .format(self.extension, "', '".join(file_names), fname))
        return fname


class Converter(object):
    """
    Base class for all Arcana data format converters

    Parameters
    ----------
    input_format : FileFormat
        The input format to convert from
    output_format : FileFormat
        The output format to convert to
    """

    requirements = []

    def __init__(self, input_format, output_format, wall_time=None,
                 mem_gb=None):
        self._input_format = input_format
        self._output_format = output_format
        self._wall_time = wall_time
        self._mem_gb = mem_gb

    def __eq__(self, other):
        return (self.input_format == self.input_format and
                self._output_format == other.output_format)

    @property
    def input_format(self):
        return self._input_format

    @property
    def output_format(self):
        return self._output_format

    @property
    def interface(self):
        # To be overridden by subclasses
        return NotImplementedError

    @property
    def input(self):
        # To be overridden by subclasses
        return NotImplementedError

    @property
    def output(self):
        # To be overridden by subclasses
        return NotImplementedError

    @property
    def mem_gb(self):
        return self._mem_gb

    @property
    def wall_time(self):
        return self._wall_time

    def __repr__(self):
        return "{}(input_format={}, output_format={})".format(
            type(self).__name__, self.input_format, self.output_format)


class IdentityConverter(Converter):

    requirements = []
    interface = IdentityInterface(['i'])
    input = 'i'
    output = 'i'
